==> chezmoi-debug-2.66.0-1-x86_64.pkg.tar.zst: usr/src/debug/chezmoi/chezmoi-2.66.0/internal/chezmoi/diff.go <==
package chezmoi

import (
	"io/fs"
	"net/http"
	"strings"
	"time"

	"github.com/go-git/go-git/v5/plumbing"
	"github.com/go-git/go-git/v5/plumbing/filemode"
	"github.com/go-git/go-git/v5/plumbing/format/diff"
	"github.com/sergi/go-diff/diffmatchpatch"
)

var gitDiffOperation = map[diffmatchpatch.Operation]diff.Operation{
	diffmatchpatch.DiffDelete: diff.Delete,
	diffmatchpatch.DiffEqual:  diff.Equal,
	diffmatchpatch.DiffInsert: diff.Add,
}

// A GitDiffChunk implements the
// github.com/go-git/go-git/v5/plumbing/format/diff.Chunk interface.
type GitDiffChunk struct {
	content   string
	operation diff.Operation
}

func (c *GitDiffChunk) Content() string      { return c.content }
func (c *GitDiffChunk) Type() diff.Operation { return c.operation }

// A GitDiffFile implements the
// github.com/go-git/go-git/v5/plumbing/format/diff.File interface.
type GitDiffFile struct {
	hash     plumbing.Hash
	fileMode filemode.FileMode
	relPath  RelPath
}

func (f *GitDiffFile) Hash() plumbing.Hash     { return f.hash }
func (f *GitDiffFile) Mode() filemode.FileMode { return f.fileMode }
func (f *GitDiffFile) Path() string            { return f.relPath.String() }

// A GitDiffFilePatch implements the
// github.com/go-git/go-git/v5/plumbing/format/diff.FilePatch interface.
type GitDiffFilePatch struct {
	binary   bool
	from, to diff.File
	chunks   []diff.Chunk
}

func (fp *GitDiffFilePatch) IsBinary() bool              { return fp.binary }
func (fp *GitDiffFilePatch) Files() (from, to diff.File) { return fp.from, fp.to }
func (fp *GitDiffFilePatch) Chunks() []diff.Chunk        { return fp.chunks }

// A GitDiffPatch implements the
// github.com/go-git/go-git/v5/plumbing/format/diff.Patch interface.
type GitDiffPatch struct {
	filePatches []diff.FilePatch
	message     string
}

func (p *GitDiffPatch) FilePatches() []diff.FilePatch { return p.filePatches }
func (p *GitDiffPatch) Message() string               { return p.message }

// DiffPatch returns a github.com/go-git/go-git/plumbing/format/diff.Patch for
// path from the given data and mode to the given data and mode.
func DiffPatch(path RelPath, fromData []byte, fromMode fs.FileMode, toData []byte, toMode fs.FileMode) (diff.Patch, error) {
	isBinary := isBinary(fromData) || isBinary(toData)

	var from diff.File
	if fromData != nil || fromMode != 0 {
		fromFileMode, err := diffFileMode(fromMode)
		if err != nil {
			return nil, err
		}
		from = &GitDiffFile{
			fileMode: fromFileMode,
			relPath:  path,
			hash:     plumbing.ComputeHash(plumbing.BlobObject, fromData),
		}
	}

	var to diff.File
	if toData != nil || toMode != 0 {
		toFileMode, err := diffFileMode(toMode)
		if err != nil {
			return nil, err
		}
		to = &GitDiffFile{
			fileMode: toFileMode,
			relPath:  path,
			hash:     plumbing.ComputeHash(plumbing.BlobObject, toData),
		}
	}

	var chunks []diff.Chunk
	if !isBinary {
		chunks = diffChunks(string(fromData), string(toData))
	}

	return &GitDiffPatch{
		filePatches: []diff.FilePatch{
			&GitDiffFilePatch{
				binary: isBinary,
				from:   from,
				to:     to,
				chunks: chunks,
			},
		},
	}, nil
}

// diffChunks returns the
// github.com/go-git/go-git/v5/plumbing/format/diff.Chunks required to transform
// from into to.
func diffChunks(from, to string) []diff.Chunk {
	dmp := diffmatchpatch.New()
	dmp.DiffTimeout = time.Second
	fromRunes, toRunes, runesToLines := dmp.DiffLinesToRunes(from, to)
	diffs := dmp.DiffCharsToLines(dmp.DiffMainRunes(fromRunes, toRunes, false), runesToLines)
	chunks := make([]diff.Chunk, len(diffs))
	for i, d := range diffs {
		chunk := &GitDiffChunk{
			content:   d.Text,
			operation: gitDiffOperation[d.Type],
		}
		chunks[i] = chunk
	}
	return chunks
}

// diffFileMode converts an io/fs.FileMode into a
// github.com/go-git/go-git/v5/plumbing/format/diff.FileMode.
func diffFileMode(mode fs.FileMode) (filemode.FileMode, error) {
	fileMode, err := filemode.NewFromOSFileMode(mode)
	if err != nil {
		return 0, err
	}
	return (fileMode &^ filemode.FileMode(fs.ModePerm)) | filemode.FileMode(mode.Perm()), nil
}

// isBinary returns true if data contains binary (non-human-readable) data.
func isBinary(data []byte) bool {
	return len(data) != 0 && !strings.HasPrefix(http.DetectContentType(data), "text/")
}

==> chezmoi-debug-2.67.1-1-x86_64.pkg.tar.zst: usr/src/debug/chezmoi/chezmoi-2.67.1/internal/chezmoi/diff.go <==
package chezmoi

import (
	"io/fs"
	"net/http"
	"strings"

	"github.com/go-git/go-git/v5/plumbing"
	"github.com/go-git/go-git/v5/plumbing/filemode"
	"github.com/go-git/go-git/v5/plumbing/format/diff"
	znkrdiff "znkr.io/diff"
	znkrtextdiff "znkr.io/diff/textdiff"
)

var gitDiffOperation = [...]diff.Operation{
	znkrdiff.Delete: diff.Delete,
	znkrdiff.Match:  diff.Equal,
	znkrdiff.Insert: diff.Add,
}

// A GitDiffChunk implements the
// github.com/go-git/go-git/v5/plumbing/format/diff.Chunk interface.
type GitDiffChunk struct {
	content   string
	operation diff.Operation
}

func (c *GitDiffChunk) Content() string      { return c.content }
func (c *GitDiffChunk) Type() diff.Operation { return c.operation }

// A GitDiffFile implements the
// github.com/go-git/go-git/v5/plumbing/format/diff.File interface.
type GitDiffFile struct {
	hash     plumbing.Hash
	fileMode filemode.FileMode
	relPath  RelPath
}

func (f *GitDiffFile) Hash() plumbing.Hash     { return f.hash }
func (f *GitDiffFile) Mode() filemode.FileMode { return f.fileMode }
func (f *GitDiffFile) Path() string            { return f.relPath.String() }

// A GitDiffFilePatch implements the
// github.com/go-git/go-git/v5/plumbing/format/diff.FilePatch interface.
type GitDiffFilePatch struct {
	binary   bool
	from, to diff.File
	chunks   []diff.Chunk
}

func (fp *GitDiffFilePatch) IsBinary() bool              { return fp.binary }
func (fp *GitDiffFilePatch) Files() (from, to diff.File) { return fp.from, fp.to }
func (fp *GitDiffFilePatch) Chunks() []diff.Chunk        { return fp.chunks }

// A GitDiffPatch implements the
// github.com/go-git/go-git/v5/plumbing/format/diff.Patch interface.
type GitDiffPatch struct {
	filePatches []diff.FilePatch
	message     string
}

func (p *GitDiffPatch) FilePatches() []diff.FilePatch { return p.filePatches }
func (p *GitDiffPatch) Message() string               { return p.message }

// DiffPatch returns a github.com/go-git/go-git/plumbing/format/diff.Patch for
// path from the given data and mode to the given data and mode.
func DiffPatch(path RelPath, fromData []byte, fromMode fs.FileMode, toData []byte, toMode fs.FileMode) (diff.Patch, error) {
	isBinary := isBinary(fromData) || isBinary(toData)

	var from diff.File
	if fromData != nil || fromMode != 0 {
		fromFileMode, err := diffFileMode(fromMode)
		if err != nil {
			return nil, err
		}
		from = &GitDiffFile{
			fileMode: fromFileMode,
			relPath:  path,
			hash:     plumbing.ComputeHash(plumbing.BlobObject, fromData),
		}
	}

	var to diff.File
	if toData != nil || toMode != 0 {
		toFileMode, err := diffFileMode(toMode)
		if err != nil {
			return nil, err
		}
		to = &GitDiffFile{
			fileMode: toFileMode,
			relPath:  path,
			hash:     plumbing.ComputeHash(plumbing.BlobObject, toData),
		}
	}

	var chunks []diff.Chunk
	if !isBinary {
		chunks = diffChunks(string(fromData), string(toData))
	}

	return &GitDiffPatch{
		filePatches: []diff.FilePatch{
			&GitDiffFilePatch{
				binary: isBinary,
				from:   from,
				to:     to,
				chunks: chunks,
			},
		},
	}, nil
}

// diffChunks returns the
// github.com/go-git/go-git/v5/plumbing/format/diff.Chunks required to transform
// from into to.
//
// Nothing documents this, but go-git seems to depend on never encountering
// consecutive edits of the same operation. This forces us to join the
// individual edits into chunks.
func diffChunks(from, to string) []diff.Chunk {
	edits := znkrtextdiff.Edits(from, to, znkrdiff.Minimal())
	if len(edits) == 0 {
		return nil
	}
	var chunks []diff.Chunk
	lastOp := edits[0].Op
	var sb strings.Builder
	for _, edit := range edits {
		if edit.Op != lastOp {
			if content := sb.String(); content != "" {
				chunks = append(chunks, &GitDiffChunk{
					content:   content,
					operation: gitDiffOperation[lastOp],
				})
			}
			lastOp = edit.Op
			sb.Reset()
		}
		sb.WriteString(edit.Line)
	}
	if content := sb.String(); content != "" {
		chunks = append(chunks, &GitDiffChunk{
			content:   content,
			operation: gitDiffOperation[lastOp],
		})
	}
	return chunks
}

// diffFileMode converts an io/fs.FileMode into a
// github.com/go-git/go-git/v5/plumbing/format/diff.FileMode.
func diffFileMode(mode fs.FileMode) (filemode.FileMode, error) {
	fileMode, err := filemode.NewFromOSFileMode(mode)
	if err != nil {
		return 0, err
	}
	return (fileMode &^ filemode.FileMode(fs.ModePerm)) | filemode.FileMode(mode.Perm()), nil
}

// isBinary returns true if data contains binary (non-human-readable) data.
func isBinary(data []byte) bool {
	return len(data) != 0 && !strings.HasPrefix(http.DetectContentType(data), "text/")
}
